@@ -42,7 +42,6 @@
 #include <wchar.h>
 #include <assert.h>
 #include <wctype.h>
-#include <iostream>
 #include <typeinfo>
 #include <sys/ioctl.h>
 #include <sys/stat.h>
@@ -187,8 +186,6 @@ static bool tick( Terminal::Framebuffer &state, Terminal::Framebuffer &new_frame
 
   if ( (!initialized)
        || (diff >= 0.02) ) {
-    display.downgrade( new_frame );
-
     std::string update = display.new_frame( initialized, state, new_frame );
     swrite( STDOUT_FILENO, update.c_str() );
     state = new_frame;
@@ -313,8 +310,6 @@ static void emulate_terminal( int fd )
 	perror( "ioctl TIOCSWINSZ" );
 	return;
       }
-    } else if ( sel.error( STDIN_FILENO ) || sel.error( fd ) ) {
-      break;
     }
 
     Terminal::Framebuffer new_frame( complete.get_fb() );
