@@ -75,7 +75,7 @@ static void usage( const char *argv0 ) {
   fprintf( stderr, "Copyright 2012 Keith Winstein <mosh-devel@mit.edu>\n" );
   fprintf( stderr, "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.\nThis is free software: you are free to change and redistribute it.\nThere is NO WARRANTY, to the extent permitted by law.\n\n" );
 
-  fprintf( stderr, "Usage: %s IP PORT\n       %s -c\n", argv0, argv0 );
+  fprintf( stderr, "Usage: %s [-# 'ARGS'] IP PORT\n       %s -c\n", argv0, argv0 );
 }
 
 static void print_colorcount( void )
@@ -107,8 +107,11 @@ int main( int argc, char *argv[] )
 
   /* Get arguments */
   int opt;
-  while ( (opt = getopt( argc, argv, "c" )) != -1 ) {
+  while ( (opt = getopt( argc, argv, "#:c" )) != -1 ) {
     switch ( opt ) {
+    case '#':
+      // Ignore the original arguments to mosh wrapper
+      break;
     case 'c':
       print_colorcount();
       exit( 0 );
@@ -163,12 +166,13 @@ int main( int argc, char *argv[] )
   /* Adopt native locale */
   set_native_locale();
 
+  bool success = false;
   try {
     STMClient client( ip, desired_port, key, predict_mode );
     client.init();
 
     try {
-      client.main();
+      success = client.main();
     } catch ( ... ) {
       client.shutdown();
       throw;
@@ -178,16 +182,19 @@ int main( int argc, char *argv[] )
   } catch ( const Network::NetworkException &e ) {
     fprintf( stderr, "Network exception: %s\r\n",
 	     e.what() );
+    success = false;
   } catch ( const Crypto::CryptoException &e ) {
     fprintf( stderr, "Crypto exception: %s\r\n",
 	     e.what() );
+    success = false;
   } catch ( const std::exception &e ) {
     fprintf( stderr, "Error: %s\r\n", e.what() );
+    success = false;
   }
 
   printf( "\n[mosh is exiting.]\n" );
 
   free( key );
 
-  return 0;
+  return !success;
 }
