@@ -36,6 +36,7 @@
 #include <errno.h>
 #include <locale.h>
 #include <string.h>
+#include <strings.h>
 #include <sstream>
 #include <termios.h>
 #include <unistd.h>
@@ -54,6 +55,7 @@
 #include <netdb.h>
 #include <time.h>
 #include <sys/stat.h>
+#include <inttypes.h>
 
 #ifdef HAVE_UTMPX_H
 #include <utmpx.h>
@@ -87,13 +89,15 @@
 #define _PATH_BSHELL "/bin/sh"
 #endif
 
-#include "networktransport.cc"
+#include "networktransport-impl.h"
 
 typedef Network::Transport< Terminal::Complete, Network::UserStream > ServerConnection;
 
 static void serve( int host_fd,
 		   Terminal::Complete &terminal,
-		   ServerConnection &network );
+		   ServerConnection &network,
+		   long network_timeout,
+		   long network_signaled_timeout );
 
 static int run_server( const char *desired_ip, const char *desired_port,
 		       const string &command_path, char *command_argv[],
@@ -101,9 +105,9 @@ static int run_server( const char *desired_ip, const char *desired_port,
 
 using namespace std;
 
-static void print_usage( const char *argv0 )
+static void print_usage( FILE *stream, const char *argv0 )
 {
-  fprintf( stderr, "Usage: %s new [-s] [-v] [-i LOCALADDR] [-p PORT[:PORT2]] [-c COLORS] [-l NAME=VALUE] [-- COMMAND...]\n", argv0 );
+  fprintf( stream, "Usage: %s new [-s] [-v] [-i LOCALADDR] [-p PORT[:PORT2]] [-c COLORS] [-l NAME=VALUE] [-- COMMAND...]\n", argv0 );
 }
 
 static void print_motd( void );
@@ -131,14 +135,14 @@ static string get_SSH_IP( void )
   const char *SSH_CONNECTION = getenv( "SSH_CONNECTION" );
   if ( !SSH_CONNECTION ) { /* Older sshds don't set this */
     fprintf( stderr, "Warning: SSH_CONNECTION not found; binding to any interface.\n" );
-    return string( "0.0.0.0" );
+    return string( "" );
   }
   istringstream ss( SSH_CONNECTION );
   string dummy, local_interface_IP;
   ss >> dummy >> dummy >> local_interface_IP;
   if ( !ss ) {
     fprintf( stderr, "Warning: Could not parse SSH_CONNECTION; binding to any interface.\n" );
-    return string( "0.0.0.0" );
+    return string( "" );
   }
 
   /* Strip IPv6 prefix. */
@@ -172,6 +176,10 @@ int main( int argc, char *argv[] )
 
   /* strip off command */
   for ( int i = 0; i < argc; i++ ) {
+    if ( 0 == strcmp( argv[ i ], "--help" ) || 0 == strcmp( argv[ i ], "-h" ) ) {
+      print_usage( stdout, argv[ 0 ] );
+      exit( 0 );
+    }
     if ( 0 == strcmp( argv[ i ], "--" ) ) { /* -- is mandatory */
       if ( i != argc - 1 ) {
 	command_argv = argv + i + 1;
@@ -186,8 +194,18 @@ int main( int argc, char *argv[] )
        && (strcmp( argv[ 1 ], "new" ) == 0) ) {
     /* new option syntax */
     int opt;
-    while ( (opt = getopt( argc - 1, argv + 1, "i:p:c:svl:" )) != -1 ) {
+    while ( (opt = getopt( argc - 1, argv + 1, "@:i:p:c:svl:" )) != -1 ) {
       switch ( opt ) {
+	/*
+	 * This undocumented option does nothing but eat its argument.
+	 * Useful in scripting where you prepend something to a
+	 * mosh-server argv, and might end up with something like
+	 * "mosh-server new -v new -c 256", now you can say
+	 * "mosh-server new -v -@ new -c 256" to discard the second
+	 * "new".
+	 */
+      case '@':
+	break;
       case 'i':
 	desired_ip = optarg;
 	break;
@@ -195,16 +213,19 @@ int main( int argc, char *argv[] )
 	desired_port = optarg;
 	break;
       case 's':
+	desired_ip = NULL;
 	desired_ip_str = get_SSH_IP();
-	desired_ip = desired_ip_str.c_str();
-	fatal_assert( desired_ip );
+	if ( !desired_ip_str.empty() ) {
+	  desired_ip = desired_ip_str.c_str();
+	  fatal_assert( desired_ip );
+	}
 	break;
       case 'c':
 	try {
 	  colors = myatoi( optarg );
 	} catch ( const CryptoException & ) {
 	  fprintf( stderr, "%s: Bad number of colors (%s)\n", argv[ 0 ], optarg );
-	  print_usage( argv[ 0 ] );
+	  print_usage( stderr, argv[ 0 ] );
 	  exit( 1 );
 	}
 	break;
@@ -215,7 +236,7 @@ int main( int argc, char *argv[] )
 	locale_vars.push_back( string( optarg ) );
 	break;
       default:
-	print_usage( argv[ 0 ] );
+	print_usage( stderr, argv[ 0 ] );
 	/* don't die on unknown options */
       }
     }
@@ -228,7 +249,7 @@ int main( int argc, char *argv[] )
     desired_ip = argv[ 1 ];
     desired_port = argv[ 2 ];
   } else {
-    print_usage( argv[ 0 ] );
+    print_usage( stderr, argv[ 0 ] );
     exit( 1 );
   }
 
@@ -236,7 +257,7 @@ int main( int argc, char *argv[] )
   int dpl, dph;
   if ( desired_port && ! Connection::parse_portrange( desired_port, dpl, dph ) ) {
     fprintf( stderr, "%s: Bad UDP port range (%s)\n", argv[ 0 ], desired_port );
-    print_usage( argv[ 0 ] );
+    print_usage( stderr, argv[ 0 ] );
     exit( 1 );
   }
 
@@ -334,6 +355,34 @@ int main( int argc, char *argv[] )
 static int run_server( const char *desired_ip, const char *desired_port,
 		       const string &command_path, char *command_argv[],
 		       const int colors, bool verbose, bool with_motd ) {
+  /* get network idle timeout */
+  long network_timeout = 0;
+  char *timeout_envar = getenv( "MOSH_SERVER_NETWORK_TMOUT" );
+  if ( timeout_envar && *timeout_envar ) {
+    errno = 0;
+    char *endptr;
+    network_timeout = strtol( timeout_envar, &endptr, 10 );
+    if ( *endptr != '\0' || ( network_timeout == 0 && errno == EINVAL ) ) {
+      fprintf( stderr, "MOSH_SERVER_NETWORK_TMOUT not a valid integer, ignoring\n" );
+    } else if ( network_timeout < 0 ) {
+      fprintf( stderr, "MOSH_SERVER_NETWORK_TMOUT is negative, ignoring\n" );
+      network_timeout = 0;
+    }
+  }
+  /* get network signaled idle timeout */
+  long network_signaled_timeout = 0;
+  char *signal_envar = getenv( "MOSH_SERVER_SIGNAL_TMOUT" );
+  if ( signal_envar && *signal_envar ) {
+    errno = 0;
+    char *endptr;
+    network_signaled_timeout = strtol( signal_envar, &endptr, 10 );
+    if ( *endptr != '\0' || ( network_signaled_timeout == 0 && errno == EINVAL ) ) {
+      fprintf( stderr, "MOSH_SERVER_SIGNAL_TMOUT not a valid integer, ignoring\n" );
+    } else if ( network_signaled_timeout < 0 ) {
+      fprintf( stderr, "MOSH_SERVER_SIGNAL_TMOUT is negative, ignoring\n" );
+      network_signaled_timeout = 0;
+    }
+  }
   /* get initial window size */
   struct winsize window_size;
   if ( ioctl( STDIN_FILENO, TIOCGWINSZ, &window_size ) < 0 ||
@@ -484,6 +533,13 @@ static int run_server( const char *desired_ip, const char *desired_port,
       warn_unattached( utmp_entry );
     }
 
+    /* Wait for parent to release us. */
+    char linebuf[81];
+    if (fgets(linebuf, sizeof linebuf, stdin) == NULL) {
+      perror( "parent signal" );
+      _exit( 1 );
+    }
+
     Crypto::reenable_dumping_core();
 
     if ( execvp( command_path.c_str(), command_argv ) < 0 ) {
@@ -499,7 +555,7 @@ static int run_server( const char *desired_ip, const char *desired_port,
 #endif
 
     try {
-      serve( master, terminal, *network );
+      serve( master, terminal, *network, network_timeout, network_signaled_timeout );
     } catch ( const Network::NetworkException &e ) {
       fprintf( stderr, "Network exception: %s\n",
 	       e.what() );
@@ -525,12 +581,16 @@ static int run_server( const char *desired_ip, const char *desired_port,
   return 0;
 }
 
-static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &network )
+static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &network, long network_timeout, long network_signaled_timeout )
 {
+  /* scale timeouts */
+  const uint64_t network_timeout_ms = static_cast<uint64_t>( network_timeout ) * 1000;
+  const uint64_t network_signaled_timeout_ms = static_cast<uint64_t>( network_signaled_timeout ) * 1000;
   /* prepare to poll for events */
   Select &sel = Select::get_instance();
   sel.add_signal( SIGTERM );
   sel.add_signal( SIGINT );
+  sel.add_signal( SIGUSR1 );
 
   uint64_t last_remote_num = network.get_remote_state_num();
 
@@ -541,16 +601,35 @@ static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &
   socklen_t saved_addr_len = 0;
   #endif
 
+  bool child_released = false;
+
   while ( 1 ) {
     try {
+      static const uint64_t timeout_if_no_client = 60000;
+      int timeout = INT_MAX;
       uint64_t now = Network::timestamp();
 
-      const int timeout_if_no_client = 60000;
-      int timeout = min( network.wait_time(), terminal.wait_time( now ) );
+      timeout = min( timeout, network.wait_time() );
+      timeout = min( timeout, terminal.wait_time( now ) );
       if ( (!network.get_remote_state_num())
 	   || network.shutdown_in_progress() ) {
         timeout = min( timeout, 5000 );
       }
+      /*
+       * The server goes completely asleep if it has no remote peer.
+       * We may want to wake up sooner.
+       */
+      if ( network_timeout_ms ) {
+	int64_t network_sleep = network_timeout_ms -
+	  ( now - network.get_latest_remote_state().timestamp );
+	if ( network_sleep < 0 ) {
+	  network_sleep = 0;
+	} else if ( network_sleep > INT_MAX ) {
+	  /* 24 days might be too soon.  That's OK. */
+	  network_sleep = INT_MAX;
+	}
+	timeout = min( timeout, static_cast<int>(network_sleep) );
+      }
 
       /* poll for events */
       sel.clear_fds();
@@ -570,6 +649,7 @@ static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &
 
       now = Network::timestamp();
       uint64_t time_since_remote_state = now - network.get_latest_remote_state().timestamp;
+      string terminal_to_host;
 
       if ( sel.read( network_fd ) ) {
 	/* packet received from the network */
@@ -579,17 +659,20 @@ static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &
 	if ( network.get_remote_state_num() != last_remote_num ) {
 	  last_remote_num = network.get_remote_state_num();
 
-	  string terminal_to_host;
 	  
 	  Network::UserStream us;
 	  us.apply_string( network.get_remote_diff() );
 	  /* apply userstream to terminal */
 	  for ( size_t i = 0; i < us.size(); i++ ) {
 	    const Parser::Action *action = us.get_action( i );
-	    terminal_to_host += terminal.act( action );
 	    if ( typeid( *action ) == typeid( Parser::Resize ) ) {
+	      /* apply only the last consecutive Resize action */
+	      while ( i < us.size() - 1 &&
+		   typeid( us.get_action( i + 1 ) ) == typeid( Parser::Resize ) ) {
+		i++;
+	      }
 	      /* tell child process of resize */
-	      const Parser::Resize *res = static_cast<const Parser::Resize *>( us.get_action( i ) );
+	      const Parser::Resize *res = static_cast<const Parser::Resize *>( action );
 	      struct winsize window_size;
 	      if ( ioctl( host_fd, TIOCGWINSZ, &window_size ) < 0 ) {
 		perror( "ioctl TIOCGWINSZ" );
@@ -602,6 +685,7 @@ static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &
 		return;
 	      }
 	    }
+	    terminal_to_host += terminal.act( action );
 	  }
 
 	  if ( !us.empty() ) {
@@ -614,11 +698,6 @@ static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &
 	    network.set_current_state( terminal );
 	  }
 	  
-	  /* write any writeback octets back to the host */
-	  if ( swrite( host_fd, terminal_to_host.c_str(), terminal_to_host.length() ) < 0 ) {
-	    break;
-	  }
-
 	  #ifdef HAVE_UTEMPTER
 	  /* update utmp entry if we have become "connected" */
 	  if ( (!connected_utmp)
@@ -645,6 +724,15 @@ static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &
 	    connected_utmp = true;
 	  }
 	  #endif
+
+	  /* Tell child to start login session. */
+	  if ( !child_released ) {
+	    if ( swrite( host_fd, "\n", 1 ) < 0) {
+	      perror( "child release" );
+	      _exit( 1 );
+	    }
+	    child_released = true;
+	  }
 	}
       }
       
@@ -661,19 +749,34 @@ static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &
         if ( bytes_read <= 0 ) {
 	  network.start_shutdown();
 	} else {
-	  string terminal_to_host = terminal.act( string( buf, bytes_read ) );
+	  terminal_to_host += terminal.act( string( buf, bytes_read ) );
 	
 	  /* update client with new state of terminal */
 	  network.set_current_state( terminal );
-
-	  /* write any writeback octets back to the host */
-	  if ( swrite( host_fd, terminal_to_host.c_str(), terminal_to_host.length() ) < 0 ) {
-	    break;
-	  }
 	}
       }
 
-      if ( sel.any_signal() ) {
+      /* write user input and terminal writeback to the host */
+      if ( swrite( host_fd, terminal_to_host.c_str(), terminal_to_host.length() ) < 0 ) {
+	break;
+      }
+
+      bool idle_shutdown = false;
+      if ( network_timeout_ms &&
+	   network_timeout_ms <= time_since_remote_state ) {
+	idle_shutdown = true;
+	fprintf( stderr, "Network idle for %llu seconds.\n", 
+		 static_cast<unsigned long long>( time_since_remote_state / 1000 ) );
+      }
+      if ( sel.signal( SIGUSR1 ) ) {
+	if ( !network_signaled_timeout_ms || network_signaled_timeout_ms <= time_since_remote_state ) {
+	  idle_shutdown = true;
+	  fprintf( stderr, "Network idle for %llu seconds when SIGUSR1 received\n",
+		   static_cast<unsigned long long>( time_since_remote_state / 1000 ) );
+	}
+      }
+
+      if ( sel.any_signal() || idle_shutdown ) {
 	/* shutdown signal */
 	if ( network.has_remote_addr() && (!network.shutdown_in_progress()) ) {
 	  network.start_shutdown();
@@ -682,16 +785,6 @@ static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &
 	}
       }
       
-      if ( sel.error( network_fd ) ) {
-	/* network problem */
-	break;
-      }
-
-      if ( (!network.shutdown_in_progress()) && sel.error( host_fd ) ) {
-	/* host problem */
-	network.start_shutdown();
-      }
-
       /* quit if our shutdown has been acknowledged */
       if ( network.shutdown_in_progress() && network.shutdown_acknowledged() ) {
 	break;
@@ -730,9 +823,9 @@ static void serve( int host_fd, Terminal::Complete &terminal, ServerConnection &
       }
 
       if ( !network.get_remote_state_num()
-           && time_since_remote_state >= uint64_t( timeout_if_no_client ) ) {
-        fprintf( stderr, "No connection within %d seconds.\n",
-                 timeout_if_no_client / 1000 );
+           && time_since_remote_state >= timeout_if_no_client ) {
+        fprintf( stderr, "No connection within %llu seconds.\n",
+                 static_cast<unsigned long long>( timeout_if_no_client / 1000 ) );
         break;
       }
 
