@@ -60,7 +60,7 @@
 #include "select.h"
 #include "timestamp.h"
 
-#include "networktransport.cc"
+#include "networktransport-impl.h"
 
 void STMClient::resume( void )
 {
@@ -121,8 +121,75 @@ void STMClient::init( void )
     overlays.set_title_prefix( wstring( L"[mosh] " ) );
   }
 
+  /* Set terminal escape key. */
+  const char *escape_key_env;
+  if ( (escape_key_env = getenv( "MOSH_ESCAPE_KEY" )) != NULL ) {
+    if ( strlen( escape_key_env ) == 1 ) {
+      escape_key = (int)escape_key_env[0];
+      if ( escape_key > 0 && escape_key < 128 ) {
+	if ( escape_key < 32 ) {
+	  /* If escape is ctrl-something, pass it with repeating the key without ctrl. */
+	  escape_pass_key = escape_key + (int)'@';
+	} else {
+	  /* If escape is something else, pass it with repeating the key itself. */
+	  escape_pass_key = escape_key;
+	}
+	if ( escape_pass_key >= 'A' && escape_pass_key <= 'Z' ) {
+	  /* If escape pass is an upper case character, define optional version
+	     as lower case of the same. */
+	  escape_pass_key2 = escape_pass_key + (int)'a' - (int)'A';
+	} else {
+	  escape_pass_key2 = escape_pass_key;
+	}
+      } else {
+	escape_key = 0x1E;
+	escape_pass_key = '^';
+	escape_pass_key2 = '^';
+      }
+    } else if ( strlen( escape_key_env ) == 0 ) {
+      escape_key = -1;
+    } else {
+      escape_key = 0x1E;
+      escape_pass_key = '^';
+      escape_pass_key2 = '^';
+    }
+  } else {
+    escape_key = 0x1E;
+    escape_pass_key = '^';
+    escape_pass_key2 = '^';
+  }
+
+  /* There are so many better ways to shoot oneself into leg than
+     setting escape key to Ctrl-C, Ctrl-D, NewLine, Ctrl-L or CarriageReturn
+     that we just won't allow that. */
+  if ( escape_key == 0x03 || escape_key == 0x04 || escape_key == 0x0A || escape_key == 0x0C || escape_key == 0x0D ) {
+    escape_key = 0x1E;
+    escape_pass_key = '^';
+    escape_pass_key2 = '^';
+  }
+
+  /* Adjust escape help differently if escape is a control character. */
+  if ( escape_key > 0 ) {
+    char escape_pass_name_buf[16];
+    char escape_key_name_buf[16];
+    snprintf(escape_pass_name_buf, sizeof escape_pass_name_buf, "\"%c\"", escape_pass_key);
+    if (escape_key < 32) {
+      snprintf(escape_key_name_buf, sizeof escape_key_name_buf, "Ctrl-%c", escape_pass_key);
+      escape_requires_lf = false;
+    } else {
+      snprintf(escape_key_name_buf, sizeof escape_key_name_buf, "\"%c\"", escape_key);
+      escape_requires_lf = true;
+    }
+    string tmp;
+    tmp = string( escape_pass_name_buf );
+    wstring escape_pass_name = std::wstring(tmp.begin(), tmp.end());
+    tmp = string( escape_key_name_buf );
+    wstring escape_key_name = std::wstring(tmp.begin(), tmp.end());
+    escape_key_help = L"Commands: Ctrl-Z suspends, \".\" quits, " + escape_pass_name + L" gives literal " + escape_key_name;
+    overlays.get_notification_engine().set_escape_key_string( tmp );
+  }
   wchar_t tmp[ 128 ];
-  swprintf( tmp, 128, L"Nothing received from server on UDP port %d.", port );
+  swprintf( tmp, 128, L"Nothing received from server on UDP port %s.", port.c_str() );
   connecting_notification = wstring( tmp );
 }
 
@@ -143,8 +210,8 @@ void STMClient::shutdown( void )
   }
 
   if ( still_connecting() ) {
-    fprintf( stderr, "\nmosh did not make a successful connection to %s:%d.\n", ip.c_str(), port );
-    fprintf( stderr, "Please verify that UDP port %d is not firewalled and can reach the server.\n\n", port );
+    fprintf( stderr, "\nmosh did not make a successful connection to %s:%s.\n", ip.c_str(), port.c_str() );
+    fprintf( stderr, "Please verify that UDP port %s is not firewalled and can reach the server.\n\n", port.c_str() );
     fprintf( stderr, "(By default, mosh uses a UDP port between 60000 and 61000. The -p option\nselects a specific UDP port number.)\n" );
   } else if ( network ) {
     if ( !clean_shutdown ) {
@@ -170,18 +237,18 @@ void STMClient::main_init( void )
   }  
 
   /* local state */
-  local_framebuffer = new Terminal::Framebuffer( window_size.ws_col, window_size.ws_row );
-  new_state = new Terminal::Framebuffer( 1, 1 );
+  local_framebuffer = Terminal::Framebuffer( window_size.ws_col, window_size.ws_row );
+  new_state = Terminal::Framebuffer( 1, 1 );
 
   /* initialize screen */
-  string init = display.new_frame( false, *local_framebuffer, *local_framebuffer );
+  string init = display.new_frame( false, local_framebuffer, local_framebuffer );
   swrite( STDOUT_FILENO, init.data(), init.size() );
 
   /* open network */
   Network::UserStream blank;
   Terminal::Complete local_terminal( window_size.ws_col, window_size.ws_row );
   network = new Network::Transport< Network::UserStream, Terminal::Complete >( blank, local_terminal,
-									       key.c_str(), ip.c_str(), port );
+									       key.c_str(), ip.c_str(), port.c_str() );
 
   network->set_send_delay( 1 ); /* minimal delay on outgoing keystrokes */
 
@@ -196,29 +263,23 @@ void STMClient::output_new_frame( void )
   }
 
   /* fetch target state */
-  *new_state = network->get_latest_remote_state().state.get_fb();
+  new_state = network->get_latest_remote_state().state.get_fb();
 
   /* apply local overlays */
-  overlays.apply( *new_state );
-
-  /* apply any mutations */
-  display.downgrade( *new_state );
+  overlays.apply( new_state );
 
   /* calculate minimal difference from where we are */
   const string diff( display.new_frame( !repaint_requested,
-					*local_framebuffer,
-					*new_state ) );
+					local_framebuffer,
+					new_state ) );
   swrite( STDOUT_FILENO, diff.data(), diff.size() );
 
   repaint_requested = false;
 
-  /* switch pointers */
-  Terminal::Framebuffer *tmp = new_state;
-  new_state = local_framebuffer;
-  local_framebuffer = tmp;
+  local_framebuffer = new_state;
 }
 
-bool STMClient::process_network_input( void )
+void STMClient::process_network_input( void )
 {
   network->recv();
   
@@ -229,8 +290,6 @@ bool STMClient::process_network_input( void )
   overlays.get_prediction_engine().set_local_frame_acked( network->get_sent_state_acked() );
   overlays.get_prediction_engine().set_send_interval( network->send_interval() );
   overlays.get_prediction_engine().set_local_frame_late_acked( network->get_latest_remote_state().state.get_echo_ack() );
-
-  return true;
 }
 
 bool STMClient::process_user_input( int fd )
@@ -253,9 +312,7 @@ bool STMClient::process_user_input( int fd )
     for ( int i = 0; i < bytes_read; i++ ) {
       char the_byte = buf[ i ];
 
-      overlays.get_prediction_engine().new_user_byte( the_byte, *local_framebuffer );
-
-      const static wstring help_message( L"Commands: Ctrl-Z suspends, \".\" quits, \"^\" gives literal Ctrl-^" );
+      overlays.get_prediction_engine().new_user_byte( the_byte, local_framebuffer );
 
       if ( quit_sequence_started ) {
 	if ( the_byte == '.' ) { /* Quit sequence is Ctrl-^ . */
@@ -266,7 +323,7 @@ bool STMClient::process_user_input( int fd )
 	  } else {
 	    return false;
 	  }
-	} else if ( the_byte == 0x1a ) { /* Suspend sequence is Ctrl-^ Ctrl-Z */
+	} else if ( the_byte == 0x1a ) { /* Suspend sequence is escape_key Ctrl-Z */
 	  /* Restore terminal and terminal-driver state */
 	  swrite( STDOUT_FILENO, display.close().c_str() );
 
@@ -283,30 +340,34 @@ bool STMClient::process_user_input( int fd )
 	  kill( 0, SIGSTOP );
 
 	  resume();
-	} else if ( the_byte == '^' ) {
-	  /* Emulation sequence to type Ctrl-^ is Ctrl-^ ^ */
-	  network->get_current_state().push_back( Parser::UserByte( 0x1E ) );
+	} else if ( (the_byte == escape_pass_key) || (the_byte == escape_pass_key2) ) {
+	  /* Emulation sequence to type escape_key is escape_key +
+	     escape_pass_key (that is escape key without Ctrl) */
+	  network->get_current_state().push_back( Parser::UserByte( escape_key ) );
 	} else {
-	  /* Ctrl-^ followed by anything other than . and ^ gets sent literally */
-	  network->get_current_state().push_back( Parser::UserByte( 0x1E ) );
+	  /* Escape key followed by anything other than . and ^ gets sent literally */
+	  network->get_current_state().push_back( Parser::UserByte( escape_key ) );
 	  network->get_current_state().push_back( Parser::UserByte( the_byte ) );	  
 	}
 
 	quit_sequence_started = false;
 
-	if ( overlays.get_notification_engine().get_notification_string() == help_message ) {
+	if ( overlays.get_notification_engine().get_notification_string() == escape_key_help ) {
 	  overlays.get_notification_engine().set_notification_string( L"" );
 	}
 
 	continue;
       }
 
-      quit_sequence_started = (the_byte == 0x1E);
+      quit_sequence_started = (escape_key > 0) && (the_byte == escape_key) && (lf_entered || (! escape_requires_lf));
       if ( quit_sequence_started ) {
-	overlays.get_notification_engine().set_notification_string( help_message, true, false );
+	lf_entered = false;
+	overlays.get_notification_engine().set_notification_string( escape_key_help, true, false );
 	continue;
       }
 
+      lf_entered = ( (the_byte == 0x0A) || (the_byte == 0x0D) ); /* LineFeed, Ctrl-J, '\n' or CarriageReturn, Ctrl-M, '\r' */
+
       if ( the_byte == 0x0C ) { /* Ctrl-L */
 	repaint_requested = true;
       }
@@ -341,7 +402,7 @@ bool STMClient::process_resize( void )
   return true;
 }
 
-void STMClient::main( void )
+bool STMClient::main( void )
 {
   /* initialize signal handling and structures */
   main_init();
@@ -387,15 +448,10 @@ void STMClient::main( void )
 	  /* we only read one socket each run */
 	  network_ready_to_read = true;
 	}
-
-	if ( sel.error( *it ) ) {
-	  /* network problem */
-	  break;
-	}
       }
 
       if ( network_ready_to_read ) {
-	if ( !process_network_input() ) { return; }
+	process_network_input();
       }
     
       if ( sel.read( STDIN_FILENO ) ) {
@@ -412,7 +468,7 @@ void STMClient::main( void )
 
       if ( sel.signal( SIGWINCH ) ) {
         /* resize */
-        if ( !process_resize() ) { return; }
+        if ( !process_resize() ) { return false; }
       }
 
       if ( sel.signal( SIGCONT ) ) {
@@ -432,16 +488,6 @@ void STMClient::main( void )
         }
       }
 
-      if ( sel.error( STDIN_FILENO ) ) {
-	/* user problem */
-	if ( !network->has_remote_addr() ) {
-	  break;
-	} else if ( !network->shutdown_in_progress() ) {
-	  overlays.get_notification_engine().set_notification_string( wstring( L"Exiting..." ), true );
-	  network->start_shutdown();
-	}
-      }
-
       /* quit if our shutdown has been acknowledged */
       if ( network->shutdown_in_progress() && network->shutdown_acknowledged() ) {
 	clean_shutdown = true;
@@ -500,10 +546,11 @@ void STMClient::main( void )
         throw;
       } else {
         wchar_t tmp[ 128 ];
-        swprintf( tmp, 128, L"Crypto exception: %s", e.text.c_str() );
+        swprintf( tmp, 128, L"Crypto exception: %s", e.what() );
         overlays.get_notification_engine().set_notification_string( wstring( tmp ) );
       }
     }
   }
+  return clean_shutdown;
 }
 
