@@ -61,15 +61,15 @@ void ConditionalOverlayCell::apply( Framebuffer &fb, uint64_t confirmed_epoch, i
 
   if ( unknown ) {
     if ( flag && ( col != fb.ds.get_width() - 1 ) ) {
-      fb.get_mutable_cell( row, col )->renditions.underlined = true;
+      fb.get_mutable_cell( row, col )->get_renditions().set_attribute(Renditions::underlined, true);
     }
     return;
   }
 
-  if ( !(*(fb.get_cell( row, col )) == replacement) ) {
+  if ( *fb.get_cell( row, col ) != replacement ) {
     *(fb.get_mutable_cell( row, col )) = replacement;
     if ( flag ) {
-      fb.get_mutable_cell( row, col )->renditions.underlined = true;
+      fb.get_mutable_cell( row, col )->get_renditions().set_attribute( Renditions::underlined, true );
     }
   }
 }
@@ -206,9 +206,9 @@ void NotificationEngine::apply( Framebuffer &fb ) const
 
   /* draw bar across top of screen */
   Cell notification_bar( 0 );
-  notification_bar.renditions.foreground_color = 37;
-  notification_bar.renditions.background_color = 44;
-  notification_bar.contents.push_back( 0x20 );
+  notification_bar.get_renditions().foreground_color = 37;
+  notification_bar.get_renditions().background_color = 44;
+  notification_bar.append( 0x20 );
 
   for ( int i = 0; i < fb.ds.get_width(); i++ ) {
     *(fb.get_mutable_cell( 0, i )) = notification_bar;
@@ -275,12 +275,12 @@ void NotificationEngine::apply( Framebuffer &fb ) const
     case 2: /* wide character */
       this_cell = fb.get_mutable_cell( 0, overlay_col );
       fb.reset_cell( this_cell );
-      this_cell->renditions.bold = true;
-      this_cell->renditions.foreground_color = 37;
-      this_cell->renditions.background_color = 44;
+      this_cell->get_renditions().set_attribute(Renditions::bold, true);
+      this_cell->get_renditions().foreground_color = 37;
+      this_cell->get_renditions().background_color = 44;
       
-      this_cell->contents.push_back( ch );
-      this_cell->width = chwidth;
+      this_cell->append( ch );
+      this_cell->set_width( chwidth );
       combining_cell = this_cell;
 
       overlay_col += chwidth;
@@ -290,14 +290,14 @@ void NotificationEngine::apply( Framebuffer &fb ) const
 	break;
       }
 
-      if ( combining_cell->contents.size() == 0 ) {
-	assert( combining_cell->width == 1 );
-	combining_cell->fallback = true;
+      if ( combining_cell->empty() ) {
+	assert( combining_cell->get_width() == 1 );
+	combining_cell->set_fallback( true );
 	overlay_col++;
       }
 
-      if ( combining_cell->contents.size() < 16 ) {
-	combining_cell->contents.push_back( ch );
+      if ( !combining_cell->full() ) {
+	combining_cell->append( ch );
       }
       break;
     case -1: /* unprintable character */
@@ -345,9 +345,9 @@ void OverlayManager::apply( Framebuffer &fb )
   title.apply( fb );
 }
 
-void TitleEngine::set_prefix( const wstring s )
+void TitleEngine::set_prefix( const wstring &s )
 {
-  prefix = deque<wchar_t>( s.begin(), s.end() );
+  prefix = Terminal::Framebuffer::title_type( s.begin(), s.end() );
 }
 
 void ConditionalOverlayRow::apply( Framebuffer &fb, uint64_t confirmed_epoch, bool flag ) const
@@ -560,11 +560,11 @@ void PredictionEngine::cull( const Framebuffer &fb )
 
 	/* match rest of row to the actual renditions */
 	{
-	  const Renditions &actual_renditions = fb.get_cell( i->row_num, j->col )->renditions;
+	  const Renditions &actual_renditions = fb.get_cell( i->row_num, j->col )->get_renditions();
 	  for ( overlay_cells_type::iterator k = j;
 		k != i->overlay_cells.end();
 		k++ ) {
-	    k->replacement.renditions = actual_renditions;
+	    k->replacement.get_renditions() = actual_renditions;
 	  }
 	}
 
@@ -666,9 +666,10 @@ void PredictionEngine::new_user_byte( char the_byte, const Framebuffer &fb )
   }
   last_byte = the_byte;
 
-  list<Parser::Action *> actions( parser.input( the_byte ) );
+  Parser::Actions actions;
+  parser.input( the_byte, actions );
 
-  for ( list<Parser::Action *>::iterator it = actions.begin();
+  for ( Parser::Actions::iterator it = actions.begin();
         it != actions.end();
         it++ ) {
     Parser::Action *act = *it;
@@ -678,7 +679,8 @@ void PredictionEngine::new_user_byte( char the_byte, const Framebuffer &fb )
 	     act->name().c_str(), act->char_present ? act->ch : L'_' );
     */
 
-    if ( typeid( *act ) == typeid( Parser::Print ) ) {
+    const std::type_info& type_act = typeid( *act );
+    if ( type_act == typeid( Parser::Print ) ) {
       /* make new prediction */
 
       init_cursor( fb );
@@ -775,21 +777,21 @@ void PredictionEngine::new_user_byte( char the_byte, const Framebuffer &fb )
 	cell.active = true;
 	cell.tentative_until_epoch = prediction_epoch;
 	cell.expire( local_frame_sent + 1, now );
-	cell.replacement.renditions = fb.ds.get_renditions();
+	cell.replacement.get_renditions() = fb.ds.get_renditions();
 
 	/* heuristic: match renditions of character to the left */
 	if ( cursor().col > 0 ) {
 	  ConditionalOverlayCell &prev_cell = the_row.overlay_cells[ cursor().col - 1 ];
 	  const Cell *prev_cell_actual = fb.get_cell( cursor().row, cursor().col - 1 );
 	  if ( prev_cell.active && (!prev_cell.unknown) ) {
-	    cell.replacement.renditions = prev_cell.replacement.renditions;
+	    cell.replacement.get_renditions() = prev_cell.replacement.get_renditions();
 	  } else {
-	    cell.replacement.renditions = prev_cell_actual->renditions;
+	    cell.replacement.get_renditions() = prev_cell_actual->get_renditions();
 	  }
 	}
 
-	cell.replacement.contents.clear();
-	cell.replacement.contents.push_back( ch );
+	cell.replacement.clear();
+	cell.replacement.append( ch );
 	cell.original_contents.push_back( *fb.get_cell( cursor().row, cursor().col ) );
 
 	/*
@@ -809,7 +811,7 @@ void PredictionEngine::new_user_byte( char the_byte, const Framebuffer &fb )
 	  newline_carriage_return( fb );
 	}
       }
-    } else if ( typeid( *act ) == typeid( Parser::Execute ) ) {
+    } else if ( type_act == typeid( Parser::Execute ) ) {
       if ( act->char_present && (act->ch == 0x0d) /* CR */ ) {
 	become_tentative();
 	newline_carriage_return( fb );
@@ -817,10 +819,10 @@ void PredictionEngine::new_user_byte( char the_byte, const Framebuffer &fb )
 	//	fprintf( stderr, "Execute 0x%x\n", act->ch );
 	become_tentative();	
       }
-    } else if ( typeid( *act ) == typeid( Parser::Esc_Dispatch ) ) {
+    } else if ( type_act == typeid( Parser::Esc_Dispatch ) ) {
       //      fprintf( stderr, "Escape sequence\n" );
       become_tentative();
-    } else if ( typeid( *act ) == typeid( Parser::CSI_Dispatch ) ) {
+    } else if ( type_act == typeid( Parser::CSI_Dispatch ) ) {
       if ( act->char_present && (act->ch == L'C') ) { /* right arrow */
 	init_cursor( fb );
 	if ( cursor().col < fb.ds.get_width() - 1 ) {
@@ -838,8 +840,6 @@ void PredictionEngine::new_user_byte( char the_byte, const Framebuffer &fb )
 	//	fprintf( stderr, "CSI sequence %lc\n", act->ch );
 	become_tentative();
       }
-    } else if ( typeid( *act ) == typeid( Parser::Clear ) ) {
-
     }
 
     delete act;
@@ -876,7 +876,7 @@ void PredictionEngine::newline_carriage_return( const Framebuffer &fb )
       j->active = true;
       j->tentative_until_epoch = prediction_epoch;
       j->expire( local_frame_sent + 1, now );
-      j->replacement.contents.clear();
+      j->replacement.clear();
     }
   } else {
     cursor().row++;
