@@ -66,30 +66,25 @@ using namespace Crypto;
 const uint64_t DIRECTION_MASK = uint64_t(1) << 63;
 const uint64_t SEQUENCE_MASK = uint64_t(-1) ^ DIRECTION_MASK;
 
-/* Read in packet from coded string */
-Packet::Packet( string coded_packet, Session *session )
-  : seq( -1 ),
-    direction( TO_SERVER ),
+/* Read in packet */
+Packet::Packet( const Message & message )
+  : seq( message.nonce.val() & SEQUENCE_MASK ),
+    direction( (message.nonce.val() & DIRECTION_MASK) ? TO_CLIENT : TO_SERVER ),
     timestamp( -1 ),
     timestamp_reply( -1 ),
     payload()
 {
-  Message message = session->decrypt( coded_packet );
-
-  direction = (message.nonce.val() & DIRECTION_MASK) ? TO_CLIENT : TO_SERVER;
-  seq = message.nonce.val() & SEQUENCE_MASK;
-
   dos_assert( message.text.size() >= 2 * sizeof( uint16_t ) );
 
-  uint16_t *data = (uint16_t *)message.text.data();
+  const uint16_t *data = (uint16_t *)message.text.data();
   timestamp = be16toh( data[ 0 ] );
   timestamp_reply = be16toh( data[ 1 ] );
 
   payload = string( message.text.begin() + 2 * sizeof( uint16_t ), message.text.end() );
 }
 
-/* Output coded string from packet */
-string Packet::tostring( Session *session )
+/* Output from packet */
+Message Packet::toMessage( void )
 {
   uint64_t direction_seq = (uint64_t( direction == TO_CLIENT ) << 63) | (seq & SEQUENCE_MASK);
 
@@ -98,10 +93,10 @@ string Packet::tostring( Session *session )
 
   string timestamps = string( (char *)ts_net, 2 * sizeof( uint16_t ) );
 
-  return session->encrypt( Message( Nonce( direction_seq ), timestamps + payload ) );
+  return Message( Nonce( direction_seq ), timestamps + payload );
 }
 
-Packet Connection::new_packet( string &s_payload )
+Packet Connection::new_packet( const string &s_payload )
 {
   uint16_t outgoing_timestamp_reply = -1;
 
@@ -114,7 +109,7 @@ Packet Connection::new_packet( string &s_payload )
     saved_timestamp_received_at = 0;
   }
 
-  Packet p( next_seq++, direction, timestamp16(), outgoing_timestamp_reply, s_payload );
+  Packet p( direction, timestamp16(), outgoing_timestamp_reply, s_payload );
 
   return p;
 }
@@ -180,7 +175,10 @@ Connection::Socket::Socket( int family )
   int tosflag = true;
   socklen_t tosoptlen = sizeof( tosflag );
   if ( setsockopt( _fd, IPPROTO_IP, IP_RECVTOS, &tosflag, tosoptlen ) < 0 ) {
-    perror( "setsockopt( IP_RECVTOS )" );
+    /* FreeBSD disallows this option on IPv6 sockets. */
+    if ( family == IPPROTO_IP ) {
+      perror( "setsockopt( IP_RECVTOS )" );
+    }
   }
 #endif
 }
@@ -203,6 +201,20 @@ const std::vector< int > Connection::fds( void ) const
   return ret;
 }
 
+void Connection::set_MTU( int family )
+{
+  switch ( family ) {
+  case AF_INET:
+    MTU = DEFAULT_IPV4_MTU - IPV4_HEADER_LEN;
+    break;
+  case AF_INET6:
+    MTU = DEFAULT_IPV6_MTU - IPV6_HEADER_LEN;
+    break;
+  default:
+    throw NetworkException( "Unknown address family", 0 );
+  }
+}
+
 class AddrInfo {
 public:
   struct addrinfo *res;
@@ -230,7 +242,6 @@ Connection::Connection( const char *desired_ip, const char *desired_port ) /* se
     key(),
     session( key ),
     direction( TO_CLIENT ),
-    next_seq( 0 ),
     saved_timestamp( -1 ),
     saved_timestamp_received_at( 0 ),
     expected_receiver_seq( 0 ),
@@ -319,6 +330,7 @@ bool Connection::try_bind( const char *addr, int port_low, int port_high )
     }
 
     if ( bind( sock(), &local_addr.sa, local_addr_len ) == 0 ) {
+      set_MTU( local_addr.sa.sa_family );
       return true;
     } else if ( i == search_high ) { /* last port to search */
       int saved_errno = errno;
@@ -350,7 +362,6 @@ Connection::Connection( const char *key_str, const char *ip, const char *port )
     key( key_str ),
     session( key ),
     direction( TO_SERVER ),
-    next_seq( 0 ),
     saved_timestamp( -1 ),
     saved_timestamp_received_at( 0 ),
     expected_receiver_seq( 0 ),
@@ -372,16 +383,18 @@ Connection::Connection( const char *key_str, const char *ip, const char *port )
   hints.ai_socktype = SOCK_DGRAM;
   hints.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV;
   AddrInfo ai( ip, port, &hints );
-  fatal_assert( ai.res->ai_addrlen <= sizeof( remote_addr ) );
+  fatal_assert( static_cast<size_t>( ai.res->ai_addrlen ) <= sizeof( remote_addr ) );
   remote_addr_len = ai.res->ai_addrlen;
   memcpy( &remote_addr.sa, ai.res->ai_addr, remote_addr_len );
 
   has_remote_addr = true;
 
   socks.push_back( Socket( remote_addr.sa.sa_family ) );
+
+  set_MTU( remote_addr.sa.sa_family );
 }
 
-void Connection::send( string s )
+void Connection::send( const string & s )
 {
   if ( !has_remote_addr ) {
     return;
@@ -389,7 +402,7 @@ void Connection::send( string s )
 
   Packet px = new_packet( s );
 
-  string p = px.tostring( &session );
+  string p = session.encrypt( px.toMessage() );
 
   ssize_t bytes_sent = sendto( sock(), p.data(), p.size(), MSG_DONTWAIT,
 			       &remote_addr.sa, remote_addr_len );
@@ -404,7 +417,7 @@ void Connection::send( string s )
     send_exception = NetworkException( "sendto", errno );
 
     if ( errno == EMSGSIZE ) {
-      MTU = 500; /* payload MTU of last resort */
+      MTU = DEFAULT_SEND_MTU; /* payload MTU of last resort */
     }
   }
 
@@ -493,7 +506,11 @@ string Connection::recv_one( int sock_to_recv, bool nonblocking )
   struct cmsghdr *ecn_hdr = CMSG_FIRSTHDR( &header );
   if ( ecn_hdr
        && (ecn_hdr->cmsg_level == IPPROTO_IP)
-       && (ecn_hdr->cmsg_type == IP_TOS) ) {
+       && ((ecn_hdr->cmsg_type == IP_TOS)
+#ifdef IP_RECVTOS
+	   || (ecn_hdr->cmsg_type == IP_RECVTOS)
+#endif
+	   )) {
     /* got one */
     uint8_t *ecn_octet_p = (uint8_t *)CMSG_DATA( ecn_hdr );
     assert( ecn_octet_p );
@@ -503,7 +520,7 @@ string Connection::recv_one( int sock_to_recv, bool nonblocking )
     }
   }
 
-  Packet p( string( msg_payload, received_len ), &session );
+  Packet p( session.decrypt( msg_payload, received_len ) );
 
   dos_assert( p.direction == (server ? TO_SERVER : TO_CLIENT) ); /* prevent malicious playback to sender */
 
@@ -629,9 +646,7 @@ uint64_t Connection::timeout( void ) const
 
 Connection::Socket::~Socket()
 {
-  if ( close( _fd ) < 0 ) {
-    throw NetworkException( "close", errno );
-  }
+  fatal_assert ( close( _fd ) == 0 );
 }
 
 Connection::Socket::Socket( const Socket & other )
